/****************************************************************************
 *
 *  Sprite -> PNG conversion, for g0tai, cos he's a big pudding
 *
 *  compile with: gcc -O4 -o sprite2png sprite2png.c -lpng
 *
 *  Contributors
 *  IMJ : Ian Jeffray : ian@jeffray.co.uk : Initial version
 *
 *  Versions
 *  0.01 : 05/01/2002 : IMJ : Initial version
 *
 ****************************************************************************/
 
#include "png.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "riscosmodes.h"
#include "riscoscolpal.h"

int main(int argc, char **argv)
{
    FILE *spritefile, *pngfile;
    int size;
    unsigned char *spritebuf, *spriteptr = NULL;
    
    char *progname, *spritefilename, *spritename, *pngfilename;
    
    /* Check arguments */
    
    if ( argc != 4 )
    {
        fprintf( stderr, "Usage: %s <spritefile> <spritename | -?> <pngfile>\n", argv[0] );
        exit(1);
    }
    
    /* Get some named args */
    
    progname = argv[0];
    spritefilename = argv[1];
    spritename = argv[2];
    pngfilename = argv[3];
    
    /* Check sprite name is sane */
    
    if( strlen( spritename ) > 12 )
    {
        fprintf( stderr, "%s: '%s' is not a valid sprite name\n", progname, spritename);
        exit(1);
    }
    
    /* Load in the whole sprite file */
    
    spritefile = fopen( spritefilename, "rb" );
    if (!spritefile) 
    {
        fprintf( stderr, "%s: Could not open sprite file '%s'\n", progname, spritefilename );
        exit(1);
    }
    
    fseek( spritefile, 0, SEEK_END );
    size = (int) ftell( spritefile );
    fseek( spritefile, 0, SEEK_SET );
    
    spritebuf = (unsigned char *) malloc( size );
    if (!spritebuf)
    {
        fprintf( stderr, "%s: Could not allocate %d bytes to load sprite file!\n", progname, size );
        fclose(spritefile);
        exit(1);
    }
    
    fread( spritebuf, size, 1, spritefile );
    fclose( spritefile );
    
    /* Quick check to see if we can create the output file */
    
    if (strcmp(spritename,"-?")!=0)
    {
        pngfile = fopen( pngfilename, "wb" );
        if (!pngfile)
        {
            fprintf( stderr, "%s: Could not open output file '%s' for writing!\n", progname, pngfilename );
        }
    }
    
    /* Now find the sprite we were asked for */
    
    {
        int sprites = *(int *)(spritebuf), s=sprites, ncols=256,reps=1;
        int width, realwidth, height, firstbit, lastbit, mode, i, xf, yf, bpp=0;
        unsigned char *spritebits, *maskbits, *palette;
        unsigned char trans[2]={255,0};
        png_color cpal[17], *colpal;

        spritebuf += (*(int *)(spritebuf+4)) - 4;
        
        if (strcmp(spritename,"-?")==0)
        {
            for(;s;s--)
            {
                spritebuf[4+12]=0;
                printf("=> %s\n", spritebuf+4);
                spritebuf += *(int *)spritebuf;
            }
            exit(0);
        }
            
        for(;s;s--)
        {
            if( strncasecmp( spritename, (char *) spritebuf+4, strlen(spritename) ) == 0 )
            {
                spriteptr = spritebuf;
                break;
            }
            spritebuf += *(int *)spritebuf;
        }
        if (!spriteptr)
        {
            /* Exact sprite name wasn't found.  We could try harder here (initial substring match)
               but for now, we'll just barf */
            fprintf( stderr, "%s: Could not locate sprite '%s' in spritefile\n", progname, spritename );
            exit(1);
        }
        
        /* By now, spriteptr points to the sprite we want to convert */
        
        width = *(int *)(spriteptr + 16);
        width += 1;
        width *= 4;   /* This is now the width in bytes, not pixels, and not whole used bytes due to left/right wastage! */
        
        height = *(int *)(spriteptr + 20);
        height +=1 ;  /* This is the height in pixels */
        
        firstbit = *(int *)(spriteptr + 24);
        lastbit = *(int *)(spriteptr + 28);
        
        spritebits = spriteptr + (*(int *)(spriteptr + 32));
        maskbits = spriteptr + (*(int *)(spriteptr + 36));
        
        mode = *(int *)(spriteptr + 40);
        palette = spriteptr + 44;
        
        if (maskbits == spritebits )
            maskbits = NULL;        /* No mask */
        
        if (palette == spritebits || palette == maskbits)
            palette = NULL;         /* No palette */
        
        /* Now throw out sprites which have nonsensical (user defined) old MODEs */
        
        if (mode > 44 && mode < 256 )
        {
            /* There is no way for any program to ever convert this sprite, as it relies on a user-editable
               mode definition module in RISC OS.  Users should never create sprites in such modes, so this
               error should(!) never happen */
            fprintf( stderr, "%s: Sprite is defined in MODE %d which is a user defined mode, so can never be converted!\n", progname, mode );
            exit(1);
        }
        
        /* Now let's be crap -- we'll only cope with a few MODEs for now! */
        
        for(i=0; modes[i].mode>=0; i++)
        {
            if (modes[i].mode == mode )
            {
                if (modes[i].xf!=1 || (modes[i].yf!=1 && modes[i].yf!=2) )
                {
                    fprintf(stderr, "%s: This program is crap -- should be extended to cope with MODE %d sprites!\n", progname, mode );
                    exit(1);
                }
                xf=modes[i].xf;
                yf=modes[i].yf;
                bpp=modes[i].bpp;
                break;
            }
        }
        
        if (bpp==0)
        {
            /* Mode wasn't in our list... bogus goings-on (or a "new format" sprite with mode descriptor instead of number */
            fprintf(stderr, "%s: This program is crap -- can't cope with sprite in MODE %d\n", progname, mode );
            exit(1);
        }
        
        realwidth = width;
        if (bpp < 8) realwidth <<= (4-(bpp==4?3:bpp));
        realwidth -= (31-lastbit)/bpp;
        realwidth -= firstbit/bpp;
        
        printf("This sprite is defined in MODE %d, xf=%d, yf=%d, bpp=%d, pixelwidth=%d, height=%d\n", mode, xf, yf, bpp, realwidth, height );
        
        /* Now we know the scaling (xf,yf) and number of colours (1<<bpp) for this sprite */
        
        if (yf==2)
          reps=2;
        
        if (bpp < 8 )
        {
            int row,col,bitmask=(1<<bpp)-1,shift=0;
            unsigned char *tmp, *inm, *ins;
            tmp = (unsigned char *) malloc(realwidth*height);
            ncols = (1<<bpp)+1;
            
            if (palette)
            {
                colpal = cpal;
                memset(&cpal[0], 0, 17*sizeof(png_color));
                for(i=0; i<(1<<bpp); i++)
                {
                    cpal[i+1].red  =palette[1];
                    cpal[i+1].green=palette[2];
                    cpal[i+1].blue =palette[3];
                    palette+=8;
                }
            }
            else
            {
                colpal = (bpp==4? colpal16 : bpp==2 ? colpal4 : colpal2);
            }
            
            ins = spritebits;
            spritebits = tmp;
            inm = maskbits;
            for(row=0; row<height; row++)
            {
                unsigned char *tmpinm = inm;
                unsigned char *tmpins = ins;
                ins += firstbit>>3;
                inm += firstbit>>3;
                
                shift=firstbit & 7;

                for(col=0; col<realwidth; col++)
                {
                    unsigned char pixel = *ins;
                    unsigned char mask = 255;
                    if (maskbits)
                    {
                        mask = *inm;
                        mask >>= shift;
                        mask &= bitmask;
                    }
                    
                    pixel >>= shift;
                    pixel &= bitmask;
                    
                    *tmp = pixel+1;
                    if (!mask)
                        *tmp = 0;
                    
                    tmp++;
                    shift += bpp;
                    
                    if (shift>7) { ins++; inm++; shift=0; }
                    
                }
                ins = tmpins + width;
                inm = tmpinm + width;
            }
            firstbit=0;
            width=realwidth;
        }
        
        /* Now we have 8bpp data to convert to PNG! */
            
        if (bpp==8) makeriscos256colpal();
        
        {
            png_infop info_ptr;
            png_structp png_ptr;
            
            png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, (png_voidp)NULL, NULL, NULL);
            if (!png_ptr)
               exit(2);

            info_ptr = png_create_info_struct(png_ptr);
            if (!info_ptr)
            {
               png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
               exit(2);
            }
/*
            if (setjmp(png_jmpbuf(png_ptr)))
            {
               png_destroy_write_struct(&png_ptr, &info_ptr);
               fclose(pngfile);
               exit(2);
            }
*/
            png_init_io(png_ptr, pngfile);

            png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);

            png_set_IHDR(png_ptr, info_ptr, realwidth, height*reps, 8 /* bpp */, PNG_COLOR_TYPE_PALETTE, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT );

            png_set_PLTE(png_ptr, info_ptr,  bpp==8?riscos256colpal:colpal, ncols );

            if (maskbits)
                png_set_tRNS(png_ptr, info_ptr, bpp==8?trans:trans+1, bpp==8?2:1, NULL );

            png_write_info(png_ptr, info_ptr);
            
            png_set_packing(png_ptr);

            for (i=0; i<height; i++)
            {
                int j;
                png_bytep row_pointer = spritebits + (firstbit>>3);
                spritebits += width;
                
                if (maskbits && (bpp>4) )
                {
                   unsigned char *maskp = maskbits + (firstbit>>3);
                   maskbits += width;
                
                   for(j=0; j<realwidth; j++)
                   {
                       if (row_pointer[j] == 1)
                           row_pointer[j] = 0;
                       
                       if (maskp[j] == 0)
                           row_pointer[j]=1;
                   }
                }
                
                for(j=0; j<reps; j++)
                  png_write_row(png_ptr, row_pointer);
            }
           
            png_write_end(png_ptr, info_ptr);

            png_destroy_write_struct(&png_ptr, &info_ptr);

        }
    }
}
